'use strict';

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _sinon = require('sinon');

var _sinon2 = _interopRequireDefault(_sinon);

var _mockFs = require('mock-fs');

var _mockFs2 = _interopRequireDefault(_mockFs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _precinct = require('../precinct');

var _precinct2 = _interopRequireDefault(_precinct);

var _rewire = require('rewire');

var _rewire2 = _interopRequireDefault(_rewire);

var _Config = require('./Config');

var _Config2 = _interopRequireDefault(_Config);

var _resolve = require('resolve');

var _resolve2 = _interopRequireDefault(_resolve);

var _typescript = require('typescript');

var _typescript2 = _interopRequireDefault(_typescript);

var _moduleDefinition = require('module-definition');

var _moduleDefinition2 = _interopRequireDefault(_moduleDefinition);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var dependencyTree = (0, _rewire2.default)('./');

// Bootstrap lazy requires


describe('dependencyTree', function () {
  this.timeout(8000);
  function testTreesForFormat(format) {
    var ext = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '.js';

    it('returns an object form of the dependency tree for a file', function () {
      var root = `${__dirname}/../../../../fixtures/dependency-tree/${format}`;
      var filename = `${root}/a${ext}`;

      var tree = dependencyTree({ filename, root });

      (0, _assert2.default)(tree instanceof Object);

      var aSubTree = tree[filename];

      _assert2.default.ok(aSubTree instanceof Object);
      var filesInSubTree = Object.keys(aSubTree);

      _assert2.default.equal(filesInSubTree.length, 2);
    });
  }

  function mockStylus() {
    (0, _mockFs2.default)({
      [__dirname + '/example/stylus']: {
        'a.styl': `
          @import "b"
          @require "c.styl"
        `,
        'b.styl': '@import "c"',
        'c.styl': ''
      }
    });
  }

  function mockSass() {
    (0, _mockFs2.default)({
      [__dirname + '/example/sass']: {
        'a.scss': `
          @import "_b";
          @import "_c.scss";
        `,
        '_b.scss': 'body { color: blue; }',
        '_c.scss': 'body { color: pink; }'
      }
    });
  }

  function mockLess() {
    (0, _mockFs2.default)({
      [__dirname + '/example/less']: {
        'a.less': `
          @import "b.css";
          @import "c.less";
        `,
        'b.css': 'body { color: blue; }',
        'c.less': 'body { color: pink; }'
      }
    });
  }

  function mockes6() {
    (0, _mockFs2.default)({
      [__dirname + '/example/es6']: {
        'a.js': `
          import b from './b';
          import c from './c';
        `,
        'b.js': 'export default function() {};',
        'c.js': 'export default function() {};',
        'jsx.js': `import c from './c';\n export default <jsx />;`,
        'foo.jsx': `import React from 'react';\n import b from 'b';\n export default <jsx />;`,
        'es7.js': `import c from './c';\n export default async function foo() {};`
      }
    });
  }

  function mockTS() {
    (0, _mockFs2.default)({
      [__dirname + '/example/ts']: {
        'a.ts': `
          import b from './b';
          import c from './c';
        `,
        'b.ts': 'export default function() {};',
        'c.ts': 'export default function() {};'
      }
    });
  }

  afterEach(function () {
    _mockFs2.default.restore();
  });

  it('returns an empty object for a non-existent filename', function () {
    (0, _mockFs2.default)({
      imaginary: {}
    });

    var root = __dirname + '/imaginary';
    var filename = root + '/notafile.js';
    var tree = dependencyTree({ filename, root });

    (0, _assert2.default)(tree instanceof Object);
    (0, _assert2.default)(!Object.keys(tree).length);
  });

  it('handles nested tree structures', function () {
    (0, _mockFs2.default)({
      [__dirname + '/extended']: {
        'a.js': `var b = require('./b');
                 var c = require('./c');`,
        'b.js': `var d = require('./d');
                 var e = require('./e');`,
        'c.js': `var f = require('./f');
                 var g = require('./g');`,
        'd.js': '',
        'e.js': '',
        'f.js': '',
        'g.js': ''
      }
    });

    var directory = __dirname + '/extended';
    var filename = directory + '/a.js';

    var tree = dependencyTree({ filename, directory });
    (0, _assert2.default)(tree[filename] instanceof Object);

    // b and c
    var subTree = tree[filename];
    _assert2.default.equal(Object.keys(subTree).length, 2);

    var bTree = subTree[directory + '/b.js'];
    var cTree = subTree[directory + '/c.js'];
    // d and e
    _assert2.default.equal(Object.keys(bTree).length, 2);
    // f ang g
    _assert2.default.equal(Object.keys(cTree).length, 2);
  });

  it('does not include files that are not real (#13)', function () {
    (0, _mockFs2.default)({
      [__dirname + '/onlyRealDeps']: {
        'a.js': 'var notReal = require("./notReal");'
      }
    });

    var directory = __dirname + '/onlyRealDeps';
    var filename = directory + '/a.js';

    var tree = dependencyTree({ filename, directory });
    var subTree = tree[filename];

    _assert2.default.ok(!Object.keys(subTree).some(function (dep) {
      return dep.indexOf('notReal') !== -1;
    }));
  });

  it('does not choke on cyclic dependencies', function () {
    (0, _mockFs2.default)({
      [__dirname + '/cyclic']: {
        'a.js': 'var b = require("./b");',
        'b.js': 'var a = require("./a");'
      }
    });

    var directory = __dirname + '/cyclic';
    var filename = directory + '/a.js';

    var spy = _sinon2.default.spy(dependencyTree, '_getDependencies');

    var tree = dependencyTree({ filename, directory });

    (0, _assert2.default)(spy.callCount === 2);
    (0, _assert2.default)(Object.keys(tree[filename]).length);

    dependencyTree._getDependencies.restore();
  });

  it('excludes Nodejs core modules by default', function () {
    var directory = __dirname + '/example/commonjs';
    var filename = directory + '/b.js';

    var tree = dependencyTree({ filename, directory });
    (0, _assert2.default)(Object.keys(tree[filename]).length === 0);
    (0, _assert2.default)(Object.keys(tree)[0].indexOf('b.js') !== -1);
  });

  it('traverses installed 3rd party node modules', function () {
    var directory = __dirname + '/example/onlyRealDeps';
    var filename = directory + '/a.js';

    var tree = dependencyTree({ filename, directory });
    var subTree = tree[filename];

    (0, _assert2.default)(Object.keys(subTree).some(function (dep) {
      return dep === require.resolve('debug');
    }));
  });

  it('returns a list of absolutely pathed files', function () {
    var directory = __dirname + '/example/commonjs';
    var filename = directory + '/b.js';

    var tree = dependencyTree({ filename, directory });

    for (var node in tree.nodes) {
      (0, _assert2.default)(node.indexOf(process.cwd()) !== -1);
    }
  });

  it('excludes duplicate modules from the tree', function () {
    (0, _mockFs2.default)({
      root: {
        // More than one module includes c
        'a.js': `import b from "b";
                 import c from "c";`,
        'b.js': 'import c from "c";',
        'c.js': 'export default 1;'
      }
    });

    var tree = dependencyTree.toList({
      filename: 'root/a.js',
      directory: 'root'
    });

    (0, _assert2.default)(tree.length === 3);
  });

  describe('when given a detective configuration', function () {
    it('passes it through to precinct', function () {
      var spy = _sinon2.default.spy(_precinct2.default, 'paperwork');
      var directory = __dirname + '/example/onlyRealDeps';
      var filename = directory + '/a.js';
      var detectiveConfig = {
        amd: {
          skipLazyLoaded: true
        }
      };

      dependencyTree({
        filename,
        directory,
        detective: detectiveConfig
      });

      _assert2.default.ok(spy.calledWith(filename, detectiveConfig));
      spy.restore();
    });
  });

  describe('when given a list to store non existent partials', function () {
    describe('and the file contains no valid partials', function () {
      it('stores the invalid partials', function () {
        (0, _mockFs2.default)({
          [__dirname + '/onlyRealDeps']: {
            'a.js': 'var notReal = require("./notReal");'
          }
        });

        var directory = __dirname + '/onlyRealDeps';
        var filename = directory + '/a.js';
        var nonExistent = [];

        var tree = dependencyTree({ filename, directory, nonExistent });

        _assert2.default.equal(nonExistent.length, 1);
        _assert2.default.equal(nonExistent[0], './notReal');
      });
    });

    describe('and the file contains all valid partials', function () {
      it('does not store anything', function () {
        (0, _mockFs2.default)({
          [__dirname + '/onlyRealDeps']: {
            'a.js': 'var b = require("./b");',
            'b.js': 'export default 1;'
          }
        });

        var directory = __dirname + '/onlyRealDeps';
        var filename = directory + '/a.js';
        var nonExistent = [];

        var tree = dependencyTree({ filename, directory, nonExistent });

        _assert2.default.equal(nonExistent.length, 0);
      });
    });

    describe('and the file contains a mix of invalid and valid partials', function () {
      it('stores the invalid ones', function () {
        (0, _mockFs2.default)({
          [__dirname + '/onlyRealDeps']: {
            'a.js': 'var b = require("./b");',
            'b.js': 'var c = require("./c"); export default 1;',
            'c.js': 'var crap = require("./notRealMan");'
          }
        });

        var directory = __dirname + '/onlyRealDeps';
        var filename = directory + '/a.js';
        var nonExistent = [];

        var tree = dependencyTree({ filename, directory, nonExistent });

        _assert2.default.equal(nonExistent.length, 1);
        _assert2.default.equal(nonExistent[0], './notRealMan');
      });
    });

    describe('and there is more than one reference to the invalid partial', function () {
      it('only includes the non-existent partial once', function () {
        (0, _mockFs2.default)({
          [__dirname + '/onlyRealDeps']: {
            'a.js': 'var b = require("./b");\nvar crap = require("./notRealMan");',
            'b.js': 'var c = require("./c"); export default 1;',
            'c.js': 'var crap = require("./notRealMan");'
          }
        });

        var directory = __dirname + '/onlyRealDeps';
        var filename = directory + '/a.js';
        var nonExistent = [];

        var tree = dependencyTree({ filename, directory, nonExistent });

        _assert2.default.equal(nonExistent.length, 1);
        _assert2.default.equal(nonExistent[0], './notRealMan');
      });
    });
  });

  describe('throws', function () {
    beforeEach(function () {
      this._directory = __dirname + '/example/commonjs';
      this._revert = dependencyTree.__set__('traverse', function () {
        return [];
      });
    });

    afterEach(function () {
      this._revert();
    });

    it('throws if the filename is missing', function () {
      _assert2.default.throws(function () {
        dependencyTree({
          filename: undefined,
          directory: this._directory
        });
      });
    });

    it('throws if the root is missing', function () {
      _assert2.default.throws(function () {
        dependencyTree({ filename });
      });
    });

    it('throws if a supplied filter is not a function', function () {
      _assert2.default.throws(function () {
        var directory = __dirname + '/example/onlyRealDeps';
        var filename = directory + '/a.js';

        var tree = dependencyTree({
          filename,
          directory,
          filter: 'foobar'
        });
      });
    });

    it('does not throw on the legacy `root` option', function () {
      _assert2.default.doesNotThrow(function () {
        var directory = __dirname + '/example/onlyRealDeps';
        var filename = directory + '/a.js';

        var tree = dependencyTree({
          filename,
          root: directory
        });
      });
    });
  });

  describe('on file error', function () {
    beforeEach(function () {
      this._directory = __dirname + '/example/commonjs';
    });

    it('does not throw', function () {
      var _this = this;

      _assert2.default.doesNotThrow(function () {
        dependencyTree({
          filename: 'foo',
          directory: _this._directory
        });
      });
    });

    it('returns no dependencies', function () {
      var tree = dependencyTree({ filename: 'foo', directory: this._directory });
      (0, _assert2.default)(!tree.length);
    });
  });

  describe('memoization (#2)', function () {
    beforeEach(function () {
      this._spy = _sinon2.default.spy(dependencyTree, '_getDependencies');
    });

    afterEach(function () {
      dependencyTree._getDependencies.restore();
    });

    it('accepts a cache object for memoization (#2)', function () {
      var filename = __dirname + '/example/amd/a.js';
      var directory = __dirname + '/example/amd';
      var cache = {};

      cache[__dirname + '/example/amd/b.js'] = [__dirname + '/example/amd/b.js', __dirname + '/example/amd/c.js'];

      var tree = dependencyTree({
        filename,
        directory,
        visited: cache
      });

      _assert2.default.equal(Object.keys(tree[filename]).length, 2);
      (0, _assert2.default)(this._spy.neverCalledWith(__dirname + '/example/amd/b.js'));
    });

    it('returns the precomputed list of a cached entry point', function () {
      var filename = __dirname + '/example/amd/a.js';
      var directory = __dirname + '/example/amd';

      var cache = {
        // Shouldn't process the first file's tree
        [filename]: []
      };

      var tree = dependencyTree({
        filename,
        directory,
        visited: cache
      });

      (0, _assert2.default)(!tree.length);
    });
  });

  describe('module formats', function () {
    describe('amd', function () {
      testTreesForFormat('amd');
    });

    describe('commonjs', function () {
      testTreesForFormat('commonjs');
    });

    describe('es6', function () {
      beforeEach(function () {
        this._directory = __dirname + '/example/es6';
        mockes6();
      });

      testTreesForFormat('es6');

      it('resolves files that have jsx', function () {
        var filename = `${this._directory}/jsx.js`;

        var _dependencyTree = dependencyTree({
          filename,
          directory: this._directory
        }),
            tree = _dependencyTree[filename];

        _assert2.default.ok(tree[`${this._directory}/c.js`]);
      });

      it('resolves files with a jsx extension', function () {
        var filename = `${this._directory}/foo.jsx`;

        var _dependencyTree2 = dependencyTree({
          filename,
          directory: this._directory
        }),
            tree = _dependencyTree2[filename];

        _assert2.default.ok(tree[`${this._directory}/b.js`]);
      });

      it('resolves files that have es7', function () {
        var filename = `${this._directory}/es7.js`;

        var _dependencyTree3 = dependencyTree({
          filename,
          directory: this._directory
        }),
            tree = _dependencyTree3[filename];

        _assert2.default.ok(tree[`${this._directory}/c.js`]);
      });
    });

    describe('sass', function () {
      beforeEach(function () {
        mockSass();
      });

      testTreesForFormat('sass', '.scss');
    });

    describe('stylus', function () {
      beforeEach(function () {
        mockStylus();
      });

      testTreesForFormat('stylus', '.styl');
    });

    describe('less', function () {
      beforeEach(function () {
        mockLess();
      });

      testTreesForFormat('less', '.less');
    });

    describe('typescript', function () {
      beforeEach(function () {
        mockTS();
      });

      testTreesForFormat('ts', '.ts');
    });
  });

  describe('toList', function () {
    function testToList(format) {
      var ext = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '.js';

      it('returns a post-order list form of the dependency tree', function () {
        var directory = __dirname + '/example/' + format;
        var filename = directory + '/a' + ext;

        var list = dependencyTree.toList({
          filename,
          directory
        });

        (0, _assert2.default)(list instanceof Array);
        (0, _assert2.default)(list.length);
      });
    }

    it('returns an empty list on a non-existent filename', function () {
      (0, _mockFs2.default)({
        imaginary: {}
      });

      var directory = __dirname + '/imaginary';
      var filename = directory + '/notafile.js';

      var list = dependencyTree.toList({
        filename,
        directory
      });

      (0, _assert2.default)(list instanceof Array);
      (0, _assert2.default)(!list.length);
    });

    it('orders the visited files by last visited', function () {
      var directory = __dirname + '/example/amd';
      var filename = directory + '/a.js';
      var list = dependencyTree.toList({
        filename,
        directory
      });

      (0, _assert2.default)(list.length === 3);
      (0, _assert2.default)(list[0] === directory + '/c.js');
      (0, _assert2.default)(list[1] === directory + '/b.js');
      (0, _assert2.default)(list[list.length - 1] === filename);
    });

    describe('module formats', function () {
      describe('amd', function () {
        testToList('amd');
      });

      describe('commonjs', function () {
        testToList('commonjs');
      });

      describe('es6', function () {
        beforeEach(function () {
          mockes6();
        });

        testToList('es6');
      });

      describe('sass', function () {
        beforeEach(function () {
          mockSass();
        });

        testToList('sass', '.scss');
      });

      describe('stylus', function () {
        beforeEach(function () {
          mockStylus();
        });

        testToList('stylus', '.styl');
      });

      describe('less', function () {
        beforeEach(function () {
          mockLess();
        });

        testToList('less', '.less');
      });

      describe('typescript', function () {
        beforeEach(function () {
          mockTS();
        });

        testToList('ts', '.ts');
      });
    });
  });

  describe('webpack', function () {
    beforeEach(function () {
      var _this2 = this;

      // Note: not mocking because webpack's resolver needs a real project with dependencies;
      // otherwise, we'd have to mock a ton of files.
      this._root = _path2.default.join(__dirname, '../');
      this._webpackConfig = this._root + '/webpack.config.js';

      this._testResolution = function (name) {
        var results = dependencyTree.toList({
          filename: `${__dirname}/example/webpack/${name}.js`,
          directory: _this2._root,
          webpackConfig: _this2._webpackConfig,
          filter: function filter(filename) {
            return filename.indexOf('filing-cabinet') !== -1;
          }
        });

        _assert2.default.ok(results.some(function (filename) {
          return filename.indexOf('node_modules/filing-cabinet') !== -1;
        }));
      };
    });

    it('resolves aliased modules', function () {
      this.timeout(5000);
      this._testResolution('aliased');
    });

    it('resolves unaliased modules', function () {
      this.timeout(5000);
      this._testResolution('unaliased');
    });
  });

  describe('requirejs', function () {
    beforeEach(function () {
      (0, _mockFs2.default)({
        root: {
          'lodizzle.js': 'define({})',
          'require.config.js': `
            requirejs.config({
              baseUrl: './',
              paths: {
                F: './lodizzle.js'
              }
            });
          `,
          'a.js': `
            define([
              'F'
            ], function(F) {

            });
          `,
          'b.js': `
            define([
              './lodizzle'
            ], function(F) {

            });
          `
        }
      });
    });

    it('resolves aliased modules', function () {
      var tree = dependencyTree({
        filename: 'root/a.js',
        directory: 'root',
        config: 'root/require.config.js'
      });

      var filename = _path2.default.resolve(process.cwd(), 'root/a.js');
      var aliasedFile = _path2.default.resolve(process.cwd(), 'root/lodizzle.js');
      _assert2.default.ok('root/lodizzle.js' in tree[filename]);
    });

    it('resolves non-aliased paths', function () {
      var tree = dependencyTree({
        filename: 'root/b.js',
        directory: 'root',
        config: 'root/require.config.js'
      });

      var filename = _path2.default.resolve(process.cwd(), 'root/b.js');
      var aliasedFile = _path2.default.resolve(process.cwd(), 'root/lodizzle.js');
      _assert2.default.ok('root/lodizzle.js' in tree[filename]);
    });
  });

  describe('when a filter function is supplied', function () {
    it('uses the filter to determine if a file should be included in the results', function () {
      var directory = __dirname + '/example/onlyRealDeps';
      var filename = directory + '/a.js';

      var tree = dependencyTree({
        filename,
        directory,
        // Skip all 3rd party deps
        filter: function filter(filePath, moduleFile) {
          _assert2.default.ok(require.resolve('debug'));
          _assert2.default.ok(moduleFile.match('test/example/onlyRealDeps/a.js'));
          return filePath.indexOf('node_modules') === -1;
        }
      });

      var subTree = tree[filename];
      _assert2.default.ok(Object.keys(tree).length);

      var has3rdPartyDep = Object.keys(subTree).some(function (dep) {
        return dep === require.resolve('debug');
      });
      _assert2.default.ok(!has3rdPartyDep);
    });
  });

  describe('when given a CJS file with lazy requires', function () {
    beforeEach(function () {
      (0, _mockFs2.default)({
        [__dirname + '/cjs']: {
          'foo.js': 'module.exports = function(bar = require("./bar")) {};',
          'bar.js': 'module.exports = 1;'
        }
      });
    });

    it('includes the lazy dependency', function () {
      var directory = __dirname + '/cjs';
      var filename = directory + '/foo.js';

      var tree = dependencyTree({ filename, directory });
      var subTree = tree[filename];

      _assert2.default.ok(`${directory}/bar.js` in subTree);
    });
  });

  describe('when given an es6 file using CJS lazy requires', function () {
    beforeEach(function () {
      (0, _mockFs2.default)({
        [__dirname + '/es6']: {
          'foo.js': 'export default function(bar = require("./bar")) {};',
          'bar.js': 'export default 1;'
        }
      });
    });

    describe('and mixedImport mode is turned on', function () {
      it('includes the lazy dependency', function () {
        var directory = __dirname + '/es6';
        var filename = directory + '/foo.js';

        var tree = dependencyTree({
          filename,
          directory,
          detective: {
            es6: {
              mixedImports: true
            }
          }
        });

        var subTree = tree[filename];

        _assert2.default.ok(`${directory}/bar.js` in subTree);
      });

      it('also works for toList', function () {
        var directory = __dirname + '/es6';
        var filename = directory + '/foo.js';

        var results = dependencyTree.toList({
          filename,
          directory,
          detective: {
            es6: {
              mixedImports: true
            }
          }
        });

        _assert2.default.equal(results[0], `${directory}/bar.js`);
        _assert2.default.equal(results[1], filename);
      });
    });

    describe('and mixedImport mode is turned off', function () {
      it('does not include the lazy dependency', function () {
        var directory = __dirname + '/es6';
        var filename = directory + '/foo.js';

        var tree = dependencyTree({
          filename,
          directory
        });

        var subTree = tree[filename];

        _assert2.default.ok(!(`${directory}/bar.js` in subTree));
      });
    });
  });

  describe('when given an es6 file using dynamic imports', function () {
    beforeEach(function () {
      (0, _mockFs2.default)({
        [__dirname + '/es6']: {
          'foo.js': 'import("./bar");',
          'bar.js': 'export default 1;'
        }
      });
    });

    it('includes the dynamic import', function () {
      var directory = __dirname + '/es6';
      var filename = directory + '/foo.js';

      var tree = dependencyTree({
        filename,
        directory
      });

      var subTree = tree[filename];

      _assert2.default.ok(!(`${directory}/bar.js` in subTree));
    });
  });

  describe('when given a CJS file with module property in package.json', function () {
    beforeEach(function () {
      (0, _mockFs2.default)({
        [__dirname + '/es6']: {
          ['module.entry.js']: 'import * as module from "module.entry"',
          ['node_modules']: {
            ['module.entry']: {
              'index.main.js': 'module.exports = function() {};',
              'index.module.js': 'module.exports = function() {};',
              'package.json': '{ "main": "index.main.js", "module": "index.module.js" }'
            }
          }
        }
      });
    });

    it('it includes the module entry as dependency', function () {
      var directory = __dirname + '/es6';
      var filename = directory + '/module.entry.js';

      var tree = dependencyTree({
        filename,
        directory,
        nodeModulesConfig: {
          entry: 'module'
        }
      });
      var subTree = tree[filename];

      _assert2.default.ok(`${directory}/node_modules/module.entry/index.module.js` in subTree);
    });
  });

  describe('Config', function () {
    describe('when cloning', function () {
      describe('and a detective config was set', function () {
        it('retains the detective config in the clone', function () {
          var detectiveConfig = {
            es6: {
              mixedImports: true
            }
          };

          var config = new _Config2.default({
            detectiveConfig,
            filename: 'foo',
            directory: 'bar'
          });

          var clone = config.clone();

          _assert2.default.deepEqual(clone.detectiveConfig, detectiveConfig);
        });
      });
    });
  });
});
//# sourceMappingURL=index.spec.js.map