'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = generateTree;

var _os = require('os');

var _os2 = _interopRequireDefault(_os);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _dependencyTree = require('./dependency-tree');

var _dependencyTree2 = _interopRequireDefault(_dependencyTree);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Check if running on Windows.
 * @type {Boolean}
 */
var isWin = _os2.default.platform() === 'win32';

/**
 * Check if path is from NPM folder
 * @param  {String} path
 * @return {Boolean}
 */
// most of the functions in this file were taken from the Madge project: https://github.com/pahen/madge
// reasons for not using Madge directly: 1) it has issues with TypeScript on Windows. 2) it has issues with tsx files

function isNpmPathFunc(path) {
  return path.indexOf('node_modules') >= 0;
}

/**
 * Sort tree.
 * @param  {Object} tree
 * @return {Object}
 */
function sort(tree) {
  return Object.keys(tree).sort().reduce(function (acc, id) {
    acc[id] = tree[id].sort();
    return acc;
  }, {});
}

/**
 * Exclude modules from tree using RegExp.
 * @param  {Object} tree
 * @param  {Array} excludeRegExp
 * @return {Object}
 */
function exclude(tree, excludeRegExp) {
  var regExpList = excludeRegExp.map(function (re) {
    return new RegExp(re);
  });

  function regExpFilter(id) {
    return regExpList.findIndex(function (regexp) {
      return regexp.test(id);
    }) < 0;
  }

  return Object.keys(tree).filter(regExpFilter).reduce(function (acc, id) {
    acc[id] = tree[id].filter(regExpFilter);
    return acc;
  }, {});
}

/**
 * Process absolute path and return a shorter one.
 * @param  {String} absPath
 * @param  {Object} cache
 * @param  {String} baseDir
 * @return {String}
 */
function processPath(absPath, cache, baseDir) {
  if (cache[absPath]) {
    return cache[absPath];
  }

  var relPath = _path2.default.relative(baseDir, absPath);

  if (isWin) {
    relPath = relPath.replace(/\\/g, '/');
  }

  cache[absPath] = relPath;

  return relPath;
}

/**
 * Convert deep tree produced by dependency-tree to a
 * shallow (one level deep) tree used by madge.
 * @param  {Object} depTree
 * @param  {Object} tree
 * @param  {Object} pathCache
 * @param  {String} baseDir
 * @return {Object}
 */
function convertTree(depTree, tree, pathCache, baseDir) {
  for (var key in depTree) {
    var id = processPath(key, pathCache, baseDir);

    if (!tree[id]) {
      tree[id] = [];

      for (var dep in depTree[key]) {
        tree[id].push(processPath(dep, pathCache, baseDir));
      }
    }

    convertTree(depTree[key], tree, pathCache, baseDir);
  }

  return tree;
}

function addRelativePathsToPathMap(pathMap, pathCache, baseDir) {
  pathMap.forEach(function (file) {
    file.relativePath = processPath(file.file, pathCache, baseDir);
    file.dependencies.forEach(function (dependency) {
      dependency.relativePath = processPath(dependency.resolvedDep, pathCache, baseDir);
    });
  });
}

/**
 * Generate the tree from the given files
 * @param  {Array} files
 * @param config
 * @return {Object}
 */
function generateTree(files, config) {
  var depTree = {};
  var visited = {};
  var nonExistent = {};
  var npmPaths = {};
  var pathCache = {};
  var pathMap = [];

  files.forEach(function (file) {
    if (visited[file]) {
      return;
    }

    var detective = config.detectiveOptions;
    var dependencyTreeResult = (0, _dependencyTree2.default)({
      filename: file,
      directory: config.baseDir,
      requireConfig: config.requireConfig,
      webpackConfig: config.webpackConfig,
      resolveConfig: config.resolveConfig,
      visited,
      filter: function filter(dependencyFilePath, traversedFilePath) {
        var dependencyFilterRes = true;
        var isNpmPath = isNpmPathFunc(dependencyFilePath);

        if (config.dependencyFilter) {
          dependencyFilterRes = config.dependencyFilter(dependencyFilePath, traversedFilePath, config.baseDir);
        }

        if (config.includeNpm && isNpmPath) {
          (npmPaths[traversedFilePath] = npmPaths[traversedFilePath] || []).push(dependencyFilePath);
        }

        return !isNpmPath && (dependencyFilterRes || dependencyFilterRes === undefined);
      },
      detective,
      nonExistent,
      pathMap
    });
    Object.assign(depTree, dependencyTreeResult);
  });

  var tree = convertTree(depTree, {}, pathCache, config.baseDir);

  var _loop = function _loop(npmKey) {
    var id = processPath(npmKey, pathCache, config.baseDir);

    npmPaths[npmKey].forEach(function (npmPath) {
      tree[id].push(processPath(npmPath, pathCache, config.baseDir));
    });
  };

  for (var npmKey in npmPaths) {
    _loop(npmKey);
  }

  if (config.excludeRegExp) {
    tree = exclude(tree, config.excludeRegExp);
  }

  addRelativePathsToPathMap(pathMap, pathCache, config.baseDir);

  return {
    tree: sort(tree),
    skipped: nonExistent,
    pathMap
  };
}
//# sourceMappingURL=generate-tree-madge.js.map